"""
@File       build_warning_scope.py
@Brief      Scope of a build warning
@Author     rajaber
@Date       03-22-2021
@copyright  Microsoft Corporation. All rights reserved.
"""
from enum import Enum

class BuildWarningScope(Enum):
    """Scope of a build warning"""
    FILE = 1
    TARGET = 2



"""
@File       xcwarnings.py
@Brief      Module that detects new build warnings
@Author     rajaber
@Date       03-22-2021
@copyright  Microsoft Corporation. All rights reserved.
"""
import argparse
import json
import os
import re
import copy

from utils import utils
from xcwarnings.build_warning import BuildWarning
from xcwarnings.build_warning_diff_report import BuildWarningDiffReport

# AUTO_TRIAGGED_TAG = 'AutoTriagged'
# FAILED_AUTO_TRIAGGED_TAG = 'NoAutoTriagged'


def get_arguments():
    """Get arguments input from the script execution
    RETURNS:
    - Arguments to use in script
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "xcode_build_output_file_path", help="Path to the xcode output file"
    )
    parser.add_argument(
        "--known_build_warnings_file_path",
        nargs=1,
        help="Full path to a file with known build warnings",
    )
    parser.add_argument(
        "--source_root", required=True, help="File path for the root of the source code"
    )
    parser.add_argument(
        "--generate_baseline",
        action="store_true",
        default=False,
        help="Whether a new baseline of known issues should be generated.",
    )
    return parser.parse_args()


def read_known_issues(
    known_build_warnings_file_path: str | None,
) -> dict[BuildWarning, int]:
    """Reads known build issues per optional configuration file.
    PARAMS:
    - known_build_warnings_file_path: Path to build warnings file, or None if no such file exists.
    RETURNS:
    - Dictionary of known build issues where the keys are the warnings of type BuildWarning,
      and the values are the counts of how many such warnings are expected in their respective files
      or project targets.
      Note: Those BuildWarning instances are not expected to contain line and column numbers because
      those tend to change as new content is added/removed.
    """

    if known_build_warnings_file_path is None:
        utils.print_info_message(
            "known_build_warnings_file_path not specified. No known warnings."
        )
        return {}

    utils.print_info_message(
        f"known_build_warnings_file_path = {known_build_warnings_file_path}\n"
    )

    with open(known_build_warnings_file_path, encoding="utf-8") as file:
        data_file = file.read()

    known_issues_list = json.loads(data_file)
    return {
        BuildWarning.from_dict(known_issue): known_issue["count"]
        for known_issue in known_issues_list
    }


def scan_line_for_warning(line: str, source_root: str) -> BuildWarning | None:
    """Scans given line from build output for a warning.
    PARAMS:
    - line: Contents of a line of a build output file (as a string)
    - source_root: File path for the root of the source code
    RETURNS:
    - BuildWarning instance or None if no warning is found in this line.
    """

    file_scope_warning_pattern = r"^(.*):(\d+):(\d+): warning: (.*)$"
    target_scope_warning_pattern = (
        r"^warning: (.*) \(in target '(.*)' from project '(.*)'\)$"
    )

    file_scope_warning_search = re.search(file_scope_warning_pattern, line)
    if file_scope_warning_search:
        file_path = file_scope_warning_search.group(1)
        line_number = int(file_scope_warning_search.group(2))
        column = int(file_scope_warning_search.group(3))
        warning_statement = file_scope_warning_search.group(4)

        if not file_path.startswith(os.path.abspath(source_root) + os.sep):
            # Exclude files not under source root
            return None

        relative_file_path = os.path.relpath(file_path, source_root)
        build_warning = BuildWarning.create_file_warning(
            relative_file_path, line_number, column, warning_statement
        )
        return build_warning

    target_scope_warning_search = re.search(target_scope_warning_pattern, line)
    if target_scope_warning_search:
        warning_statement = target_scope_warning_search.group(1)
        target = target_scope_warning_search.group(2)
        project = target_scope_warning_search.group(3)
        build_warning = BuildWarning.create_target_warning(
            target, project, warning_statement
        )
        return build_warning

    return None


def scan_for_warnings(output_file_path: str, source_root: str) -> [BuildWarning]:
    """Scans the given file for all build warnings
    PARAMS:
    - output_file_path: Path to a file containing logs from xcode build command
    RETURNS:
    - source_root: File path for the root of the source code
    - An array of BuildWarning instances found. Empty array if none is found.
      Note: there may be duplicate build warnings if output file contains multiple lines refering to
      the same warning messages. They will show up as duplicate entries in the returned array.
    """
    with open(output_file_path, encoding="utf-8") as file:
        data_file = file.readlines()

    warnings: [BuildWarning] = []

    for line in data_file:
        warning = scan_line_for_warning(line, source_root)
        if warning:
            warnings.append(warning)

    return warnings


def get_warnings(output_file_path: str, source_root: str) -> set[BuildWarning]:
    """Gets all warnings found in given output file, deduped, and aggregated on the file level.
    PARAMS:
    - output_file_path: Path to a file containing logs from xcode build command
    - source_root: File path for the root of the source code
    RETURNS:
    - An array of BuildWarning instances found. Empty array if none is found.
      Note: Duplicate build warnings in the output file with multiple lines refering to
      the same warning messages are deduped and will count as one error.
    """
    utils.print_info_message(
        f"Extracting warnings from build output file: {output_file_path}"
    )

    warnings = scan_for_warnings(output_file_path, source_root)
    warnings_deduped = set(warnings)
    utils.print_info_message(
        f"Total number of warning instances found in build log file: {len(warnings_deduped)}."
    )

    return warnings_deduped


def get_aggregated_warnings(warnings: set[BuildWarning]) -> dict[BuildWarning, int]:
    """Aggregates warnings so that all warnings in one file or target, have one entry in the output dictionary
    along with a count of how many times it occured
    PARAMS:
    - warnings: Array of warnings found
    RETURNS:
    - A dictionary of BuildWarning instances found. Each entry corresponds to one file or project/target.
    """

    warnings_copy = copy.deepcopy(warnings)

    for warning in warnings_copy:
        # Clear line number and column so that we can aggregate at the file level
        warning.line_number = None
        warning.column = None

    warning_dict: dict[BuildWarning, int] = {}
    for warning in warnings_copy:
        warning_dict[warning] = warning_dict.get(warning, 0) + 1

    return warning_dict


def generate_baseline(
    output_file_path: str, output_known_issues_file: str, source_root
) -> None:
    """Gets baseline for all known issues identified in the output file path, and writes them
    to a known issues file at the given path.
    PARAMS:
    - output_file_path: Path to a file containing logs from xcode build command
    - output_known_issues_file: Path to a new file that this method will write, summarizing issues
    discovered in output file path.
    - source_root: File path for the root of the source code
    RETURNS:
    - Writes known issues in JSON format at output_known_issues_file. Does not return value.
    """
    utils.print_info_message("Generating baseline 📈")
    warnings = get_warnings(output_file_path, source_root)
    warnings_with_count = get_aggregated_warnings(warnings)
    supressions = []

    for unique_warning, count in warnings_with_count.items():
        supression_entry = unique_warning.to_dict()
        supression_entry["count"] = count
        supressions.append(supression_entry)

    known_issues_json = json.dumps(supressions, indent=4)

    with open(output_known_issues_file, "w", encoding="utf-8") as text_file:
        text_file.write(known_issues_json)

    utils.print_success_message(
        f"Baseline generated at file: {output_known_issues_file}"
    )


def analyze(
    xcode_build_output_file_path: str,
    known_build_warnings_file_path: str | None,
    source_root: str,
) -> BuildWarningDiffReport:
    """Compares warnings present in the given output file with known warnings in
       known_build_warnings_file_path.
    PARAMS:
    - xcode_build_output_file_path: Path to a file containing logs from xcode build command
    - output_known_issues_file: Path to file with known build issues, or None if no known issues.
    - source_root: File path for the root of the source code
    RETURNS:
    - An instance of BuildWarningDiffReport
    """
    known_issues = read_known_issues(known_build_warnings_file_path)
    current_warnings = get_warnings(xcode_build_output_file_path, source_root)
    aggregated_warnings = get_aggregated_warnings(current_warnings)
    return BuildWarningDiffReport.calculate_diff_report(
        known_issues, aggregated_warnings, current_warnings
    )


def main():
    """Parses xcode build output file, and confirms there are no new warning errors.
    When --generate_baseline argument is passed, a baseline is generated with warnings found
    in given output log.
    PARAMS:
    - xcode_build_output_file_path: Path to a file containing logs from xcode build command
    - output_known_issues_file: Path to file with known build issues, or None if no known issues.
    RETURNS:
    - An instance of BuildWarningDiffReport
    """

    args = get_arguments()

    utils.print_info_message("Hello. Staring Build Warnings Detector ⚠️ ")

    if args.generate_baseline:
        if not args.known_build_warnings_file_path:
            raise ValueError(
                "known_build_warnings_file_path is required if "
                "generate_baseline option is specified."
            )

        generate_baseline(
            output_file_path=args.xcode_build_output_file_path,
            output_known_issues_file=args.known_build_warnings_file_path[0],
            source_root=args.source_root,
        )
        return 0

    utils.print_info_message("Analyzing current warnings against known issues if any🔬.")
    diff_report = analyze(
        xcode_build_output_file_path=args.xcode_build_output_file_path,
        known_build_warnings_file_path=args.known_build_warnings_file_path[0],
        source_root=args.source_root,
    )
    utils.print_info_message("Build Warnings Gate finished running. 🔥")
    return 0 if diff_report.success() else 1


if __name__ == "__main__":
    main()



"""
@File       build_warning_diff.py
@Brief      Build warning diff based on analyzing current warnings against known issues.
@Author     rajaber
@Date       03-22-2021
@copyright  Microsoft Corporation. All rights reserved.
"""
from xcwarnings.build_warning import BuildWarning


class BuildWarningDiff:
    """Holds counts for a given build warning: expected and actual counts"""

    def __init__(
        self, build_warning: BuildWarning, known_count: int, new_count: int
    ) -> None:
        self.build_warning = build_warning
        self.known_count = known_count
        self.new_count = new_count

    def __repr__(self) -> str:
        return f"({self.build_warning}, known count = {self.known_count}, new count = {self.new_count})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BuildWarningDiff):
            return (
                (self.build_warning == other.build_warning)
                and (self.known_count == other.known_count)
                and (self.new_count == other.new_count)
            )

        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.build_warning, self.known_count, self.new_count))



"""
@File       build_warning_diff_report.py
@Brief      Build warning diff report based on analyzing current warnings against known issues.
@Author     rajaber
@Date       03-22-2021
@copyright  Microsoft Corporation. All rights reserved.
"""
from utils import utils
from xcwarnings.build_warning import BuildWarning
from xcwarnings.build_warning_diff import BuildWarningDiff


class BuildWarningDiffReport:
    """Contains a diff result between known warnings and current warnings observed in log file"""

    # pylint: disable = too-many-arguments
    def __init__(
        self,
        diff_warning_got_worse: list[BuildWarningDiff],
        diff_new_warning: list[BuildWarningDiff],
        diff_warning_fixed: list[BuildWarningDiff],
        diff_warning_improved: list[BuildWarningDiff],
        all_warnings: set[BuildWarning],
    ) -> None:
        self.diff_warning_got_worse = diff_warning_got_worse
        self.diff_new_warning = diff_new_warning
        self.diff_warning_fixed = diff_warning_fixed
        self.diff_warning_improved = diff_warning_improved
        self.all_warnings = all_warnings

    # pylint: enable = too-many-arguments

    def success(self) -> bool:
        """Returns whether the warnings quality gate has passed, i.e no new regressions and
        known build warnings file is up to date.

        Returns:
            bool: True if warnings quality gate passed; otherwise False.
        """
        return (
            len(self.diff_warning_got_worse)
            + len(self.diff_new_warning)
            + len(self.diff_warning_fixed)
            + len(self.diff_warning_improved)
            == 0
        )

    def __repr__(self) -> str:
        return (
            f"(worse: {self.diff_warning_got_worse}, new: {self.diff_new_warning}, "
            + f"fixed: {self.diff_warning_fixed}, improved: {self.diff_warning_improved}"
        )

    @staticmethod
    def calculate_diff_report(
        known_issues: dict[BuildWarning, int],
        current_issues: dict[BuildWarning, int],
        all_warnings: set[BuildWarning],
    ) -> "__class__":
        """Generates a report of diff'ed warnings between known issues and current issues found.
        PARAMS:
        - known_issues: Known issues dictionary, where keys are BuildWarning, and value is integer,
                        which is count.
        - current_issues: Current issues dictionary, where keys are BuildWarning, and value is integer,
                        which is count.
        - all_warnings: All warnings encountered in the current build
        RETURNS:
        - An instance of BuildWarningDiffReport of warings that got worse, new warnings or warnings that
        seem to have been fixed or improved.
        """
        diff_warning_got_worse = []
        diff_new_warning = []
        diff_warning_fixed = []
        diff_warning_improved = []

        for current_issue_warning, current_count in current_issues.items():
            current_count = current_issues[current_issue_warning]
            if current_issue_warning in known_issues:
                known_count = known_issues[current_issue_warning]
                if current_count > known_count:
                    utils.print_error_message("Error: The following warning got worse.")
                    utils.print_error_message(
                        f"Known count = {known_count}, Current count = {current_count}."
                    )
                    utils.print_info_message(f"{current_issue_warning}\n")
                    diff_warning_got_worse.append(
                        BuildWarningDiff(
                            current_issue_warning, known_count, current_count
                        )
                    )
            else:
                utils.print_error_message(
                    "Error: Following issue is not a known issue."
                )
                utils.print_info_message(f"{current_issue_warning}\n")
                diff_new_warning.append(
                    BuildWarningDiff(current_issue_warning, 0, current_count)
                )

        for known_issue, known_count in known_issues.items():
            if known_issue not in current_issues:
                utils.print_error_message(
                    "Error: known issue below no longer shows up in the build."
                )
                utils.print_error_message("Please remove it from known issues.")
                utils.print_info_message(str(known_issue))
                diff_warning_fixed.append(BuildWarningDiff(known_issue, known_count, 0))
            else:
                current_count = current_issues[known_issue]
                if current_count < known_count:
                    utils.print_error_message(
                        "Error: Issue has higher count in config than in build."
                    )
                    utils.print_error_message("Please adjust count.")
                    utils.print_error_message(
                        f"Expected count from known issues file: {known_count}. \
                        Found count in the log: {current_count}."
                    )
                    utils.print_info_message(f"{known_issue }\n")
                    diff_warning_improved.append(
                        BuildWarningDiff(known_issue, known_count, current_count)
                    )

        return BuildWarningDiffReport(
            diff_warning_got_worse,
            diff_new_warning,
            diff_warning_fixed,
            diff_warning_improved,
            all_warnings,
        )



"""
@File       build_warning.py
@Brief      An Xcode build warning
@Author     rajaber
@Date       03-22-2021
@copyright  Microsoft Corporation. All rights reserved.
"""
from xcwarnings.build_warning_scope import BuildWarningScope


class BuildWarning:
    # pylint: disable = too-many-arguments
    """BuildWarning contains metadata about a warning in an xcode build output file"""

    def __init__(
        self,
        file_path: str | None,
        line_number: str | None,
        column: str | None,
        target: str | None,
        project: str | None,
        warning_statement: str,
    ) -> None:
        self.file_path = file_path
        self.line_number = line_number
        self.column = column
        self.target = target
        self.project = project
        self.warning_statement = warning_statement
        if self.file_path is not None:
            self.scope = BuildWarningScope.FILE
        else:
            self.scope = BuildWarningScope.TARGET

    # pylint: enable = too-many-arguments

    def __repr__(self) -> str:
        if self.scope == BuildWarningScope.FILE:
            return (
                f"Warning in file: {self.file_path}:{self.line_number}:{self.column}. "
                + f"Warning: {self.warning_statement}"
            )

        if self.scope == BuildWarningScope.TARGET:
            return f"Warning in target: {self.target}. Project: {self.project}. Warning: {self.warning_statement}"

        raise ValueError(f"Unexpected scope: {self.scope}")

    def __eq__(self, other: object):
        """Checks if currrent instance is eual to the other instance

        Args:
            other (object): object to compare to current instance for equality

        Returns:
            bool: True if objects are equal; otherwise, false.
        """
        if isinstance(other, BuildWarning):
            return (
                (self.file_path == other.file_path)
                and (self.warning_statement == other.warning_statement)
                and (self.project == other.project)
                and (self.target == other.target)
                and (self.line_number == other.line_number)
                and (self.column == other.column)
            )

        return False

    def __ne__(self, other: object) -> bool:
        """Checks if currrent instance is not equal to the other instance

        Args:
            other (object): object to compare to current instance for equality

        Returns:
            bool: True if objects are not equal; otherwise, true.
        """
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """Returns hash of current instance

        Returns:
            str: Hash of BuildWarning instance
        """
        return hash(self.__repr__())

    def to_dict(self) -> dict[str, object]:
        """Serializes BuildWarning instance into a dictionary."""
        warning_dict: dict[str, object] = {}
        warning_dict["warning"] = self.warning_statement
        if self.scope == BuildWarningScope.FILE:
            warning_dict["file_path"] = self.file_path
            if self.line_number:
                warning_dict["line_number"] = self.line_number
            if self.column:
                warning_dict["column"] = self.column
        elif self.scope == BuildWarningScope.TARGET:
            warning_dict["target"] = self.target
            warning_dict["project"] = self.project
        else:
            raise ValueError(f"Unexpected scope: {self.scope}")
        return warning_dict

    @staticmethod
    def create_file_warning(
        file_path: str,
        line_number: int | None,
        column: int | None,
        warning_statement: str,
    ) -> "BuildWarning":
        """Creates a file level warning

        Args:
            file_path (str): Path to the file where the warning occurs, relative to the repo root.
            line_number (int): Line number where the warning occurs
            column (int): Column within the line where the warning occurs
            warning_statement (str): Warning statement

        Returns:
            BuildWarning: a file level warning
        """
        return BuildWarning(
            file_path, line_number, column, None, None, warning_statement
        )

    @staticmethod
    def create_target_warning(
        target: str, project: str, warning_statement: str
    ) -> "BuildWarning":
        """Creates a target level warning

        Args:
            target (str): Target where the warning occurs
            project (str): Project where the warning occurs
            warning_statement (str): Warning statement

        Returns:
            BuildWarning: a target level warning
        """
        return BuildWarning(None, None, None, target, project, warning_statement)

    @staticmethod
    def from_dict(input_dict: dict[str, object]) -> "BuildWarning":
        """Creates an instance of BuildWarning from given dictinoary.

        Args:
            input_dict: dictionary serialization of BuildWarning

        Returns:
            BuildWarning: An instance of build warning
        """
        return BuildWarning(
            input_dict.get("file_path"),
            input_dict.get("line_number"),
            input_dict.get("column"),
            input_dict.get("target"),
            input_dict.get("project"),
            input_dict["warning"],
        )
